/* $Workfile: OsaLinuxRtl.c $   $Revision: 36 $    $Modtime: 12/12/08 2:08p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 1998-2003, All Rights Reserved

*******************************************************************************/

#include "../OsaRtl.h"

#ifdef DAVINCI
#include <linux/vmalloc.h>
#include <linux/sched.h>

#endif

// Headers that cannot be included in C++ code even with macro fixes
// hence we need to provide C functions here for which we give access to the C++ parts
// using our header (OsaLinuxRtl.h), that is written so that it compiles both in C and C++
// (with extern "C" linkage in C++, of course)

#include <asm/io.h>   //for 2.4, this has to be here, otherwise loading driver fails
#include <linux/interrupt.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/tqueue.h>
#else
#include <linux/workqueue.h>
#endif
#include <linux/mm.h>
#include <linux/mman.h>
#include <linux/highmem.h>


#if 0
#include <linux/iobuf.h>
#endif

#include "OsaLinuxCpp.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#include <linux/poll.h>
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#else
static spinlock_t cli_lock = SPIN_LOCK_UNLOCKED;
#endif


static spinlock_t ms_accessInterlocked;

static OsaBool32  ms_bUse1601TypeTimeStamp = OSA_TRUE;

#if 0 // Only for debug
#ifdef OSA_DPRINTF
#undef OSA_DPRINTF

#define OSA_DPRINTF(pars)     { printk("%s(%d) : ", __FILE__, __LINE__); printk pars; }

  #warning OSA_DPRINTF was redefined!
#endif
#endif


/******************************************************************************
******************************************************************************/
void OsaInitLinuxRtl(void)
{
  ms_accessInterlocked = ms_accessInterlocked;
  spin_lock_init(&ms_accessInterlocked);

  ms_bUse1601TypeTimeStamp = OSA_TRUE;
}


/******************************************************************************
******************************************************************************/
int osalinux_wait_event_timeout(
  OsaUint32 *pEventSet,
  void **ppWaitQueue,
  OsaUint32 nWaitMaxMs
)
{
#ifndef DAVINCI // PCS_TEST :: Not able to build davinci release using this code.
                // Need to fix this issue!

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
  struct wait_queue sWaitQueue;
#else
	DECLARE_WAITQUEUE(sWaitQueue, current);
#endif
  int rc = 0;

//  OSA_DPRINTF(("OsaLinux: osalinux_wait_event_timeout \n"));

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
  sWaitQueue.task = current;
  add_wait_queue((struct wait_queue **) ppWaitQueue, &sWaitQueue);
#else
  add_wait_queue((wait_queue_head_t *) *ppWaitQueue, &sWaitQueue);
#endif

  for (;;) {
    current->state = TASK_INTERRUPTIBLE;
    if (*pEventSet) {
      break;
    }
    if (!signal_pending(current)) {
      int timeout = nWaitMaxMs / (1000 / HZ);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
      if(timeout < 0) timeout = MAX_SCHEDULE_TIMEOUT;
#endif
      if (!schedule_timeout(timeout)) {
        rc = OSA_RES_TIMEOUT;  // Timeout
        OSA_DPRINTF(("OsaLinux: OSA_RES_TIMEOUT\n"));
        break;
      }
      continue;
    }
    rc = OSA_RES_FAILED;
    OSA_DPRINTF(("OsaLinux: OSA_RES_TIMEOUT\n"));
    break;
  }
  current->state = TASK_RUNNING;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
  remove_wait_queue((struct wait_queue **) ppWaitQueue, &sWaitQueue);
#else
  remove_wait_queue((wait_queue_head_t *) *ppWaitQueue, &sWaitQueue);
#endif

  return rc;
#else
  // Davinci build
  return 0;
#endif
}

/******************************************************************************
******************************************************************************/
void *osalinux_kmalloc(size_t size)
{
#if 0 // kmalloc() can only allocate 128k on intel platforms. We need more.
  if (bAtomic) {
    return kmalloc(size, GFP_ATOMIC);
  } else {
    return kmalloc(size, GFP_KERNEL);
  }
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  /* FIXME: this should not be different from 2.6 */
  if ( OsaGetCurrentExeLevel() != OSA_EXE_LEVEL_CONTROL ) {
    return kmalloc(size, GFP_ATOMIC);
  } else {
    return vmalloc(size);
  }
#else
  if (in_interrupt() || irqs_disabled()) return kmalloc(size, GFP_ATOMIC);
  else return vmalloc(size);
#endif
}


/******************************************************************************
******************************************************************************/
void osalinux_kfree(void *ptr)
{
#if 0 // kmalloc() can only allocate 128k on intel platforms. We need more.
  kfree(ptr);
#else
  if ((unsigned int) ptr >= VMALLOC_START && (unsigned int) ptr < VMALLOC_END) {
    vfree(ptr);
  } else {
    kfree(ptr);
  }
#endif
}

/******************************************************************************
******************************************************************************/
void osalinux_wake_up(void **wait_queue)
{
#ifndef DAVINCI // PCS_TEST :: Not able to build davinci release using this code.
                // Need to fix this issue!
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
  wake_up((struct wait_queue **) wait_queue);
#else
  wake_up((wait_queue_head_t *) *wait_queue);
#endif
#endif
}



/******************************************************************************
******************************************************************************/
void osalinux_poll_wait(struct file * filp, void **wait_address, struct poll_table_struct *p)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
  poll_wait(filp, (struct wait_queue **) wait_address, p);
#else
  poll_wait(filp, (wait_queue_head_t *) *wait_address, p);
#endif
}

/******************************************************************************
******************************************************************************/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
static void 
osalinux_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
 OsaLinuxInterruptCallback(irq, dev_id, regs);
}
#else
static irqreturn_t 
osalinux_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
 OsaLinuxInterruptCallback(irq, dev_id, regs);

 return IRQ_HANDLED;
}
#endif
int osalinux_request_irq(unsigned int irq,  void *dev_id)
{
  return request_irq(irq, osalinux_interrupt, SA_INTERRUPT | SA_SHIRQ, "OSA_Device", dev_id);
}


/******************************************************************************
******************************************************************************/
void osalinux_free_irq(unsigned int irq, void *dev_id)
{
  free_irq(irq, dev_id);
}


/******************************************************************************
******************************************************************************/
void *osalinux_ioremap(unsigned long offset, unsigned long size)
{
  return ioremap(offset, size);
}


/******************************************************************************
******************************************************************************/
void osalinux_iounmap(void *addr)
{
  iounmap(addr);
}

/******************************************************************************
******************************************************************************/
wchar_t *wcscpy (wchar_t *dest, const wchar_t *source)
{
  wchar_t *d = dest;

  while (*source) {
    *(dest++) = *(source++);
  }

  *dest = 0;

  return d;
}


/******************************************************************************
******************************************************************************/
size_t wcslen(const wchar_t *psz)
{
  size_t r = 0;

  while (*(psz++)) {
    //OSA_DPRINTF(("OsaLinux: %c", (char) *str));
    r++;
  }

  //OSA_DPRINTF(("\n"));
  return r;
}

#if 0
/******************************************************************************
******************************************************************************/
wchar_t *wcscat (wchar_t *dest, const wchar_t *source)
{

  while (*dest)
    dest++;

  while (*source) {
    *(dest++) = *(source++);
  }

  *dest = 0x0;

  return dest;
}
#endif


/******************************************************************************
******************************************************************************/
wchar_t *wcsncpy(wchar_t *dest, const wchar_t *source, size_t count)
{
  wchar_t *d = dest;

  while (*source && count--) {
    *(dest++) = *(source++);
  }

  *dest = 0;

  return d;
}


/******************************************************************************

  @comm We rely on the fact that all 7 bit ASCII chars have the same numeric
  representation in the Unicode char set. We only handle conversion of
  these "code 127 or less" type Unicode characters, stored in wchar_t strings,
  to ASCII stored in char strings.

******************************************************************************/
char *wcsztosz(char *strDest, const wchar_t *strSource)
{
  while (*strSource) {
    *(strDest++) = (char) *(strSource++);
  }

  *strDest = '\0';

  return strDest; // Returning a pointer to the end allows easy for concatenation
}


/******************************************************************************
******************************************************************************/
wchar_t towupper (wchar_t c)
{
  // Only handles US English
  if (c >= L'a' && c <= L'z') {
    return c - (L'a' - L'A');
  }
  return c;
}


/******************************************************************************
******************************************************************************/
char *_strrev(char *szBuf)
{
  char chr;
  char *pszStart = szBuf;
  char *pszEnd = szBuf + strlen(szBuf) - 1;

  while (pszStart < pszEnd) {
    chr = *pszStart;
    *pszStart = *pszEnd;
    *pszEnd = chr;

    pszEnd--; pszStart++;
  }

  return szBuf;
}


/*******************************************************************************
*******************************************************************************/
OsaUint32 OsaIrpGetIoCode(IN OsaIrpP pIrp)
{
  return pIrp->nIoCode;
}


/******************************************************************************
******************************************************************************/
OsaVoidP OsaIrpGetInputBuffer(IN OsaIrpP pIrp)
{
  return pIrp->pInputBuffer;
}


/******************************************************************************
******************************************************************************/
OsaVoidP OsaIrpGetOutputBuffer(IN OsaIrpP pIrp)
{
  return pIrp->pOutputBuffer;
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaIrpGetOwnerId(IN OsaIrpP pIrp)
{
  return pIrp->nOwnerId;
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaIrpGetInputBufferSize(IN OsaIrpP pIrp)
{
  return pIrp->cbInput;
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaIrpGetOutputBufferCapacity(IN OsaIrpP pIrp)
{
  return pIrp->cbOutputCapacity;
}


/******************************************************************************
******************************************************************************/
OsaResultE  OsaIrpComplete(IN OsaIrpP pIrp, IN OsaResultE eResult, IN OsaUint32 cbOutput)
{
  pIrp->result = eResult;
  pIrp->cbOutput = cbOutput;
  return eResult;
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaReadReg32(OsaUint32 pRegister)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  return readl(pRegister);
#else
  const volatile void __iomem *addr;
  addr = (const volatile void __iomem *) pRegister;
  return readl(addr);
#endif
}


/******************************************************************************
******************************************************************************/
OsaUint16 OsaReadReg16(OsaUint32 pRegister)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  return readw(pRegister);
#else
  const volatile void __iomem *addr;
  addr = (const volatile void __iomem *) pRegister;
  return readw(addr);
#endif
}


/******************************************************************************
******************************************************************************/
OsaUint8 OsaReadReg8 (OsaUint32 pRegister)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  return readb(pRegister);
#else
  const volatile void __iomem *addr;
  addr = (const volatile void __iomem *) pRegister;
  return readb(addr);
#endif
}


/******************************************************************************
******************************************************************************/
void OsaWriteReg32(OsaUint32 pRegister, OsaUint32 nValue)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  writel(nValue, pRegister);
#else
  volatile void __iomem *addr;
  addr = (volatile void __iomem *) pRegister;
  writel(nValue, addr);
#endif
}


/******************************************************************************
******************************************************************************/
void OsaWriteReg16(OsaUint32 pRegister, OsaUint16 nValue)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  writew(nValue, pRegister);
#else
  volatile void __iomem *addr;
  addr = (volatile void __iomem *) pRegister;
  writew(nValue, addr);
#endif
}


/******************************************************************************
******************************************************************************/
void OsaWriteReg8 (OsaUint32 pRegister, OsaUint8  nValue)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  writeb(nValue, pRegister);
#else
  volatile void __iomem *addr;
  addr = (volatile void __iomem *) pRegister;
  writeb(nValue, addr);
#endif
}


/******************************************************************************
******************************************************************************/
void OsaReadRegBuffer8  (OsaUint32 pRegister, OsaUint8  *pBuffer, OsaUint32 coCount)
{
  OsaUint32 coReads;
  for (coReads = 0; coReads < coCount; coReads++, pBuffer++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    *pBuffer = readb(pRegister);
#else
    *pBuffer = OsaReadReg8(pRegister);
#endif
  }
}


/******************************************************************************
******************************************************************************/
void OsaWriteRegBuffer8 (OsaUint32 pRegister, OsaUint8  *pBuffer, OsaUint32 coCount)
{
  OsaUint32 coWrites;
  for (coWrites = 0; coWrites < coCount; coWrites++, pBuffer++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    writeb(*pBuffer, pRegister);
#else
    OsaWriteReg8((OsaUint32) pRegister,  (OsaUint8) *pBuffer);
#endif
  }
}


/******************************************************************************
******************************************************************************/
OsaUint32 OsaReadPort32(OsaUint32 pPort)
{
  return inl(pPort);
}


/******************************************************************************
******************************************************************************/
OsaUint16 OsaReadPort16(OsaUint32 pPort)
{
  return inw(pPort);
}


/******************************************************************************
******************************************************************************/
OsaUint8 OsaReadPort8 (OsaUint32 pPort)
{
  return inb(pPort);
}


/******************************************************************************
******************************************************************************/
void OsaWritePort32(OsaUint32 pPort, OsaUint32 nValue)
{
  outl(nValue, pPort);
}


/******************************************************************************
******************************************************************************/
void OsaWritePort16(OsaUint32 pPort, OsaUint16 nValue)
{
  outw(nValue, pPort);
}


/******************************************************************************
******************************************************************************/
void OsaWritePort8 (OsaUint32 pPort, OsaUint8  nValue)
{
  outb(nValue, pPort);
}


/******************************************************************************
******************************************************************************/
void OsaReadPortBuffer8 (OsaUint32 pPort, OsaUint8  *pBuffer, OsaUint32 coCount)
{
  insb(pPort, pBuffer, coCount);
}


/******************************************************************************
******************************************************************************/
void OsaWritePortBuffer8(OsaUint32 pPort, OsaUint8  *pBuffer, OsaUint32 coCount)
{
  outsb(pPort, pBuffer, coCount);
}


/******************************************************************************
  @devnote Note that the Osa Interlocked functions are only guaranteed to be
  interlocked in relation to eachother.
******************************************************************************/
OsaUint8 OsaInterlockedModifyRegBits8(OsaUint32 pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits)
{
  OsaUint8 nValue;
  __u32 nSavedFlags;

  OSA_ASSERT(!(spin_is_locked(&ms_accessInterlocked)));
  spin_lock_irqsave(&ms_accessInterlocked, nSavedFlags);

  nValue = OsaReadReg8(pRegister); // Modified for __LINUX_EL__ build

  // Clear the bits we modify
  nValue &= ~nBitsToModify;
  nValue |= nValueOfBits;

#if DBG
  // Make sure the value doesn't modify bits it shouldn't
  OSA_ASSERT((nValueOfBits & ~nBitsToModify) == 0);
#endif

  OsaWriteReg8(pRegister, nValue); // Modified for __LINUX_EL__ build

  spin_unlock_irqrestore((spinlock_t *) &ms_accessInterlocked, 
                         (unsigned long) nSavedFlags);

  return nValue;
}


/******************************************************************************
  @devnote Note that the Osa Interlocked functions are only guaranteed to be
  interlocked in relation to eachother.
******************************************************************************/
OsaUint8 OsaInterlockedModifyPortBits8(OsaUint32 pRegister, OsaUint8 nBitsToModify, OsaUint8 nValueOfBits)
{
  OsaUint8 nValue;
  __u32 nSavedFlags;

  OSA_ASSERT(!(spin_is_locked(&ms_accessInterlocked)));
  spin_lock_irqsave(&ms_accessInterlocked, nSavedFlags);
  nValue = inb(pRegister);

  // Clear the bits we modify
  nValue &= ~nBitsToModify;
  nValue |= nValueOfBits;

#if DBG
  // Make sure the value doesn't modify bits it shouldn't
  OSA_ASSERT((nValueOfBits & ~nBitsToModify) == 0);
#endif

  outb(nValue, pRegister);

  spin_unlock_irqrestore((spinlock_t *) &ms_accessInterlocked, 
                         (unsigned long) nSavedFlags);

  return nValue;
}


/******************************************************************************
******************************************************************************/
void OsaInterlockedIncrement(IN OUT OsaInterlockedIntP pValue)
{
  atomic_inc(pValue);
}


/******************************************************************************
******************************************************************************/
OsaInt32 OsaInterlockedDecrement(IN OUT OsaInterlockedIntP pValue)
{
  atomic_dec(pValue);
  return pValue->counter;
  //return 0;
}


/******************************************************************************
    Suspends execution for a specified time (in microseconds).
******************************************************************************/
void OsaDelayThreadMicroSecs(
  OsaInt64 nMicroSecs     // @parm Time to suspend the execution of the current thread
)
{
  OSA_ASSERT(nMicroSecs >= 0);

  if (nMicroSecs <= 1000*MAX_UDELAY_MS) {
    udelay(nMicroSecs);  // Simply busy wait
  } else if (nMicroSecs <= 1000000/HZ) {
    // longer wait time bet less than a clock tick
    mdelay(((OsaUint32) nMicroSecs) / 1000);   // use mdelay for full msecs (udelay can overflow)
    udelay(((OsaUint32) nMicroSecs) % 1000);
  } else {
    // Long wait times > 1 clock tick, we can use the scheduler
    // Note the tricks to avoid 64 bit division and modulo, which gcc cannot handle without emitting __divdi and __moddi
    // which are defned in libgcc.a

    OsaUint32 highPart = (OsaUint32) (nMicroSecs >> 32);
    OsaUint32 lowPart  = (OsaUint32) (nMicroSecs & 0xFFFFFFFF);

    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(lowPart / (1000000L / HZ));

    while (highPart--) {
      current->state = TASK_INTERRUPTIBLE;
      schedule_timeout(0xFFFFFFFF / (1000000L / HZ));
    }
  }

}


/******************************************************************************
  Suspends execution for a specified time (in microseconds).
  Can be used from all OSA_EXE_LEVELs, but must not be used for delays
  longer than 50 microsecs (i.e. "hardware"-type delays)
******************************************************************************/
void OsaStallExecutionMicroSecs(
  IN OsaUint32 nMicroSecs       // @parm Time to suspend the execution of the processor
)
{
  OSA_ASSERT(nMicroSecs < 50);    // Must be less than 50 because otherwise we will miss a clock tick.
  udelay(nMicroSecs);            // Simply busy wait
}


/******************************************************************************

  @devnote See Rubini pp. 396

******************************************************************************/
OsaExeLevel OsaGetCurrentExeLevel()
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
  if (in_interrupt()) {
    return OSA_EXE_LEVEL_SYSTEM;
  } else if (in_softirq()) {
    return OSA_EXE_LEVEL_SYSTEM; // Same for both
  } else {
    return OSA_EXE_LEVEL_CONTROL;
  }
#else
  if (in_interrupt()) {
    return OSA_EXE_LEVEL_SYSTEM;
  } else {
    return OSA_EXE_LEVEL_CONTROL;
  }
#endif
}

/******************************************************************************
******************************************************************************/
OsaUint32 OsaGetSystemTimeTickFrequency(void)
{
  // In Linux we only support one tick frequency (100ns ticks)
  return 10000000;
}

/******************************************************************************
******************************************************************************/
void OsaSetSystemTimeSource(OsaBool32 bUse1601TypeTimeStamp)
{
  // Not used in Linux
  ms_bUse1601TypeTimeStamp = bUse1601TypeTimeStamp;
}


/******************************************************************************
  Microsecond resolution under Linux 2.0 and greater on Intel.

  This function should return 100 nanosecond increments since absolute
  time since Jan 1, 1601 but now it does it since midnight only

******************************************************************************/
void OsaQuerySystemTime(OsaDateTime *pDateTime)
{
  struct timeval tv;
  do_gettimeofday (&tv);
  pDateTime->nUint64 = tv.tv_sec;
  pDateTime->nUint64 *= 10000000LL;
  pDateTime->nUint64 += tv.tv_usec * 10;

  // do_gettimeofday() only gets time since Jan 1, 1970, so a compensation of 
  // time between Jan. 1, 1601 and Jan. 1, 1970 is added, it is 100 nanosesonds based  
  pDateTime->nUint64 += (OsaUint64) 116444736000000000LL;
}


/******************************************************************************
  100th of a second resolution (typically)
******************************************************************************/
OsaUint32 OsaQueryClockMs(void)
{
  int i;

  if (HZ < 1000) {
    return jiffies * (1000/HZ);
  } else {
    i = HZ/1000;
    return jiffies / i;
  }
}


/******************************************************************************
    Return bits (n) starting from bit position p, right adjusted.

    Bit position 0 is at the right end, bit 31 is at the left end
    -----------------------------------------------------------------
      3 3 2 2 2 2 2 2 2 2 2 2 1 1 1 1 1 1 1 1 1 1
      1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
    -----------------------------------------------------------------
******************************************************************************/
OsaUint32 getbits(OsaUint32 x, int p, int n)
{
  if( (x == 1) && (n == 32) )
  {
    return 1;
  }
  else if ( ( (x >> 31) == 1) && (n == 32) )
  {
    return x;
  }
  else
  {
    return (x >> (p+1-n)) & ~(~0 << n);
  }
}



/******************************************************************************
  Unsigned 64-bit divided by unsigned 32-bit

  Utilize bitwise operator to perform division
******************************************************************************/
OsaUint64 OsaInt64div32(OsaUint64 divident, OsaUint32 divider)
{
  int i = 1, k = 0, flag = 0;
  OsaDateTime tempDivident, tempDivider, temp, result, tempRes, med;

  temp.nUint64    = 0LL;
  med.nUint64     = 0LL;
  result.nUint64  = 0LL;
  tempRes.nUint64 = 0LL;

  tempDivident.nUint64 = divident;
  tempDivider.nUint64  = divider;

  // check dividor
  if( tempDivider.nUint64 == 1 )
  {
    return tempDivident.nUint64;
  }
  else if( tempDivider.nUint64 == 0 )
  {
    return 0;
  }

  // check divident whether it is 64-bit or not
  if( tempDivident.sUint32.nHigh != 0)
    flag = 1;


  if( flag == 0 ) // 32-bit
  {
    result.nUint64 = tempDivident.sUint32.nLow / tempDivider.sUint32.nLow;
  }
  else            //  64-bit
  {
    int dividentHighMsbPos  = 0;
    int dividentHighBitSize = 0;
    int dividentLowMsbPos   = 31;
    int dividerBitSize = 0;

    // get High section of MSB's position of dividend
    while( (flag = getbits(tempDivident.sUint32.nHigh, 31, i++) & 1) != 1 )
    {
      if( flag == 1 )
        break;
    }
    dividentHighMsbPos = 31 - (i - 1) + 1;
    dividentHighBitSize = dividentHighMsbPos + 1;

    // get low section of MSB's position of divider
    while( (getbits(tempDivider.sUint32.nLow, 31, k++) & 1) != 1  );
    dividerBitSize = (31 - (k - 1) + 1) + 1;


    // when bit size of High is larger than the size of divider
    while( dividentHighBitSize > dividerBitSize )
    {
      k = 0;

      temp.nUint64 = 0LL;
      temp.sUint32.nHigh = getbits(tempDivident.sUint32.nHigh, dividentHighMsbPos, dividerBitSize);

      if(temp.sUint32.nHigh < tempDivider.sUint32.nLow)
      {
        k++;
        temp.sUint32.nHigh = getbits(tempDivident.sUint32.nHigh, dividentHighMsbPos, dividerBitSize + k);
      }

      dividentHighMsbPos--;
      dividentHighBitSize = dividentHighMsbPos + 1;

      // romove bits that already been used
      tempDivident.sUint32.nHigh &= ~(~0 << (dividentHighBitSize + 1 - dividerBitSize - k) );

      // get remainder
      temp.sUint32.nHigh -= tempDivider.sUint32.nLow;
      temp.sUint32.nHigh = ( temp.sUint32.nHigh << ( dividentHighBitSize + 1 - dividerBitSize - k ) );

      // add remainder to dividend
      tempDivident.sUint32.nHigh |= (getbits(temp.sUint32.nHigh, dividentHighMsbPos, dividentHighBitSize));

      // take result
      result.sUint32.nHigh |= (1 << ( dividentHighBitSize + 1 - dividerBitSize - k ) );

      // get High section of MSB's position of dividend -- since dividentHighBitSize is changed
      i = 1;
      while( (flag = getbits(tempDivident.sUint32.nHigh, 31, i++) & 1) != 1 )
      {
        if( (i == 33) || (flag == 1) )
          break;
      }
      dividentHighMsbPos = 31 - (i - 1) + 1;
      dividentHighBitSize = dividentHighMsbPos + 1;

    } // while

    tempRes.nUint64 = result.nUint64;


    // get High section of MSB's position of dividend
    i = 1;
    while( (flag = getbits(tempDivident.sUint32.nHigh, 31, i++) & 1) != 1 )
    {
      if( (i == 33) || (flag == 1) )
        break;
    }

    if ( ( i == 33 ) && ( flag != 1 ) )   // no High bit remained
    {
      dividentHighMsbPos = -1;
      dividentHighBitSize = dividentHighMsbPos + 1;
    }
    else
    {
      dividentHighMsbPos = 31 - (i - 1) + 1;
      dividentHighBitSize = dividentHighMsbPos + 1;
    }


    // when bit size of High is less than or equal to the size of divider
    while( (dividentHighBitSize <= dividerBitSize) && (dividentHighBitSize > 0) )
    {
      temp.nUint64 = getbits(tempDivident.sUint32.nHigh, dividentHighMsbPos, dividentHighBitSize);

      if( temp.nUint64 == tempDivider.sUint32.nLow )
      {
        dividentHighMsbPos = dividentHighMsbPos - dividerBitSize;
        dividentHighBitSize = dividentHighMsbPos + 1;

        tempDivident.sUint32.nHigh &= 0;
        result.sUint32.nHigh = 1;
      }
      else if( temp.nUint64 < tempDivider.sUint32.nLow )
      {
        k = 0;

        // get bits which will be subtracted by divider
        if( dividerBitSize == dividentHighBitSize )
        {
          temp.nUint64 <<= 1;
          med.nUint64 = getbits(tempDivident.sUint32.nLow, dividentLowMsbPos, (dividerBitSize - dividentHighBitSize + 1));
        }
        else
        {
          temp.nUint64 = (temp.nUint64 << (dividerBitSize - dividentHighBitSize) );
          med.nUint64 = getbits(tempDivident.sUint32.nLow, dividentLowMsbPos, (dividerBitSize - dividentHighBitSize));
        }

        med.nUint64 |= temp.nUint64;

        if(med.nUint64 < tempDivider.sUint32.nLow)
        {
          k++;

          temp.nUint64 <<= 1;
          med.nUint64 = getbits(tempDivident.sUint32.nLow, dividentLowMsbPos, (dividerBitSize - dividentHighBitSize) + 1);
          med.nUint64 |= temp.nUint64;
        }

        dividentHighMsbPos--;
        dividentHighBitSize = dividentHighMsbPos + 1;

        // remove High bits that are used
        tempDivident.sUint32.nHigh &= 0;


        if( dividerBitSize == (dividentHighBitSize + 1) ) // +1 is because dividentnHighMsbPos--
        {
          tempDivident.sUint32.nLow &= ~(~0 << (31 - (dividerBitSize - ( dividentHighBitSize + 1)) ));
        }
        else
        {
          tempDivident.sUint32.nLow &= ~(~0 << (32 - (dividerBitSize - ( dividentHighBitSize + 1)) - k ) );
        }

        med.nUint64 -= tempDivider.sUint32.nLow;

        if( dividerBitSize == (dividentHighBitSize + 1) )
          med.nUint64 = ( med.nUint64 << (31 - (dividerBitSize - ( dividentHighBitSize + 1))) );
        else
          med.nUint64 = ( med.nUint64 << (32 - (dividerBitSize - ( dividentHighBitSize + 1)) - k ) );

        // add remainder to dividend
        tempDivident.sUint32.nLow  |= med.sUint32.nLow;
        tempDivident.sUint32.nHigh |= med.sUint32.nHigh;

        // take result
        if( dividerBitSize == (dividentHighBitSize + 1) )
          result.sUint32.nLow |= (1 << (31 - (dividerBitSize - ( dividentHighBitSize + 1))) );
        else
          result.sUint32.nLow |= (1 << (32 - (dividerBitSize - ( dividentHighBitSize + 1)) - k) );


        if(dividentHighBitSize <= 0)
          break;
      }
      else if( temp.nUint64 > tempDivider.sUint32.nLow )
      {
        dividentHighMsbPos--;
        dividentHighBitSize = dividentHighMsbPos + 1;

        // romove bits that already used
        tempDivident.sUint32.nHigh &= 0;

        temp.nUint64 -= tempDivider.sUint32.nLow;
        temp.nUint64 = ( temp.nUint64 << 32 );

        // add remainder to dividend
        tempDivident.sUint32.nHigh |= (getbits(temp.sUint32.nHigh, dividentHighMsbPos, dividentHighBitSize));

        // take result
        result.sUint32.nHigh |= 1;
      }

      i = 1;
      while( (flag = getbits(tempDivident.sUint32.nHigh, 31, i++) & 1) != 1 )
      {
        if( ( i == 33 ) || ( flag == 1 ) )
          break;
      }

      if ( ( i == 33 ) && ( flag != 1 ) )
      {
        dividentHighMsbPos = -1;
        dividentHighBitSize = dividentHighMsbPos + 1;
      }
      else
      {
        dividentHighMsbPos = 31 - (i - 1) + 1;
        dividentHighBitSize = dividentHighMsbPos + 1;
      }

    } // while divident bit size > divider bit size

    // 32-bit remainder
    temp.nUint64 = tempDivident.sUint32.nLow / tempDivider.sUint32.nLow;

    result.nUint64 |= tempRes.nUint64;
    result.nUint64 |= temp.nUint64;

  } //else if -- 64-bit

  return result.nUint64;
}

/******************************************************************************
******************************************************************************/
OsaVoidP OsaLockUserMemory(IN OUT OsaUserMemoryDesc *pOsaUMD, IN OsaVoidP pUserBuffer, IN OsaUint32 uiUserbufferSize)
{
  unsigned long nDataNotCopied;
  pOsaUMD->hMemory = NULL;
  pOsaUMD->pLockedBuffer = NULL;

  if ((! pOsaUMD) || (! pUserBuffer) || (uiUserbufferSize == 0)) {
    OSA_EPRINTF(("OsaLockUserMemory: Invalid Parameter!  pOsaUMD:0x08%x  pBuffer[%d]:0x%08x\n",
        (OsaUint32) pOsaUMD,
        (OsaUint32) pUserBuffer,
        uiUserbufferSize));
    return NULL;
  }

  pOsaUMD->pLockedBuffer = kmalloc(uiUserbufferSize, GFP_KERNEL);
  if (!pOsaUMD->pLockedBuffer) {
    OSA_EPRINTF(("OsaLockUserMemory: Out of memory!\n"));
    return NULL;
  }

  nDataNotCopied =   copy_from_user(pOsaUMD->pLockedBuffer, pUserBuffer, uiUserbufferSize);
  if(nDataNotCopied != 0) {
    OSA_EPRINTF(("OsaLockUserMemory: Failed to copy all data %d\n", (int) nDataNotCopied));
  }

  pOsaUMD->iUserBufferSize  = uiUserbufferSize;
  pOsaUMD->pUserBuffer      = pUserBuffer;

  return pOsaUMD->pLockedBuffer;
}


/******************************************************************************
******************************************************************************/
OsaResultE OsaUnlockUserMemory(IN OUT OsaUserMemoryDesc *pOsaUMD)
{
  if (! pOsaUMD) {
    OSA_EPRINTF(("OsaUnlockUserMemory: Invalid Parameter!  pOsaUMD:0x08%x\n", (OsaUint32) pOsaUMD));
    return OSA_RES_INVALID_PARAMETER;
  }

  if (pOsaUMD->pLockedBuffer) {
    unsigned long nDataNotCopied;
    nDataNotCopied = copy_to_user(pOsaUMD->pUserBuffer, pOsaUMD->pLockedBuffer, pOsaUMD->iUserBufferSize);
    if(nDataNotCopied != 0) {
      OSA_EPRINTF(("OsaUnlockUserMemory: Failed to copy all data %d\n", (int) nDataNotCopied));
    }
    kfree(pOsaUMD->pLockedBuffer);
  }

  memset(pOsaUMD, 0, sizeof(*pOsaUMD));
  return OSA_RES_SUCCESS;
}

/******************************************************************************
******************************************************************************/
void *osalinux_alloc_dma_common_buf(void *pPciDev, size_t size, unsigned int *pbus_addr)
{
#ifndef DAVINCI // PCS_TEST :: Not able to build davinci release using this code.
                // PCI bus is not used. Need to fix this issue if PCI is needed?
  return pci_alloc_consistent((struct pci_dev *) pPciDev, size, (dma_addr_t *) pbus_addr);
#else
  return(0);
#endif
}

/******************************************************************************
******************************************************************************/
void osalinux_free_dma_common_buf(void *pPciDev, size_t size, unsigned int log_addr, unsigned int bus_addr)
{
#ifndef DAVINCI // PCS_TEST :: Not able to build davinci release using this code.
                // PCI bus is not used. Need to fix this issue if PCI is needed?
  pci_free_consistent((struct pci_dev *) pPciDev, size, (void *) log_addr, (dma_addr_t) bus_addr);
#endif
}

/******************************************************************************
******************************************************************************/
struct OsaLinuxTaskQueue {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  struct tq_struct m_tq_struct;   // Linux OS task queue data
#else
  struct work_struct m_tq_struct;
#endif
  void (*pfnDeferredBoardNotificationCallback)(void *, void *, void *);
  void *internalDeviceRep;
  void *pIrp;
  void *pContext;
};

static void osalinux_call_deferred_board_notification(void *pContext)
{
  struct OsaLinuxTaskQueue *pTaskQueue = (struct OsaLinuxTaskQueue *) pContext;

  (*pTaskQueue->pfnDeferredBoardNotificationCallback)(pTaskQueue->internalDeviceRep, pTaskQueue->pIrp, pTaskQueue->pContext);

  kfree(pTaskQueue);
}

void osalinux_request_deferred_board_notification(
  void (*pfnDeferredBoardNotificationCallback)(void *, void *, void *),
  void *internalDeviceRep,
  void *pIrp,
  void *pContext
)
{
  struct OsaLinuxTaskQueue *pTaskQueue;

  pTaskQueue = (struct OsaLinuxTaskQueue *) kmalloc(sizeof(struct OsaLinuxTaskQueue), GFP_ATOMIC);
  if (pTaskQueue) {
    memset(pTaskQueue, 0, sizeof(*pTaskQueue));
    pTaskQueue->pfnDeferredBoardNotificationCallback = pfnDeferredBoardNotificationCallback;
    pTaskQueue->internalDeviceRep   = internalDeviceRep;
    pTaskQueue->pIrp                = pIrp;
    pTaskQueue->pContext            = pContext;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    pTaskQueue->m_tq_struct.routine = osalinux_call_deferred_board_notification;
    pTaskQueue->m_tq_struct.data    = pTaskQueue;
    queue_task(&pTaskQueue->m_tq_struct, &tq_immediate);
    mark_bh(IMMEDIATE_BH);
#else
    INIT_WORK(&pTaskQueue->m_tq_struct, osalinux_call_deferred_board_notification, pTaskQueue);
    schedule_work(&pTaskQueue->m_tq_struct);
#endif
  } else {
    OSA_EPRINTF(("osalinux_request_deferred_board_notification:: out of memory!\n"));
  }
}

/******************************************************************************
******************************************************************************/
struct OsaLinuxDpcTaskQueue {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
  struct tq_struct m_tq_struct;   // Linux OS task queue data
#else
  struct work_struct m_tq_struct;
#endif
  void (*pfnDpc)(void *, void *, void *, void *);
  void *pDpcContext;
  void *pContext;
  void *pArg1;
  void *pArg2;
};

static void osalinux_call_dpc(void *pContext)
{
  struct OsaLinuxDpcTaskQueue *pDpcTaskQueue = (struct OsaLinuxDpcTaskQueue *) pContext;

  (*pDpcTaskQueue->pfnDpc)(pDpcTaskQueue->pDpcContext, pDpcTaskQueue->pContext, pDpcTaskQueue->pArg1, pDpcTaskQueue->pArg2);
  kfree(pDpcTaskQueue);
}


OsaBool32 osalinux_request_delayed_dpc(
  void (*pfnDpc)(void *, void *, void *, void *),
  void *pDpcContext,
  void *pContext,
  void *pArg1,
  void *pArg2,
  unsigned long dueTimeInMicroSeconds
)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
{
  // We no longer support kernels versions before 2.6.x
  OSA_EPRINTF(("osalinux_request_delayed_dpc:: Not supported in this kenel version!\n"));

  OSA_DPRINTF(("osalinux_request_delayed_dpc:: calling osalinux_request_dpc no delay\n"));
  osalinux_request_dpc(pfnDpc, pDpcContext, pContext, pArg1, pArg2);
  bResult = OSA_FALSE;
}
#else
{
  unsigned long nOneMiliSec;
  unsigned long nDelay;
  OsaBool32 bResult = OSA_TRUE;

  struct OsaLinuxDpcTaskQueue *pDpcTaskQueue;


  //printk("\n");
  //printk("osalinux_request_dpc\n");

  pDpcTaskQueue = (struct OsaLinuxDpcTaskQueue *) kmalloc(sizeof(struct OsaLinuxDpcTaskQueue), GFP_ATOMIC);
  if (pDpcTaskQueue) {
    memset(pDpcTaskQueue, 0, sizeof(*pDpcTaskQueue));
    pDpcTaskQueue->pfnDpc       = pfnDpc;
    pDpcTaskQueue->pDpcContext  = pDpcContext;
    pDpcTaskQueue->pContext     = pContext;
    pDpcTaskQueue->pArg1        = pArg1;
    pDpcTaskQueue->pArg2        = pArg2;

    nOneMiliSec = OsaQueryClockMs();

    if(dueTimeInMicroSeconds/1000 > 1) {
      // Just make it mili seconds
      nDelay = dueTimeInMicroSeconds/1000;
    }
    else {
      // Less then a mSec
      nDelay = 1;
    }


    INIT_WORK(&pDpcTaskQueue->m_tq_struct, osalinux_call_dpc, pDpcTaskQueue);
    schedule_delayed_work(&pDpcTaskQueue->m_tq_struct, nDelay);

  } else {
    OSA_EPRINTF(("osalinux_request_dpc:: out of memory!\n"));
    bResult = OSA_FALSE;
  }

  return bResult;
}
#endif


void osalinux_request_dpc(
  void (*pfnDpc)(void *, void *, void *, void *),
  void *pDpcContext,
  void *pContext,
  void *pArg1,
  void *pArg2
)
{
  struct OsaLinuxDpcTaskQueue *pDpcTaskQueue;

  pDpcTaskQueue = (struct OsaLinuxDpcTaskQueue *) kmalloc(sizeof(struct OsaLinuxDpcTaskQueue), GFP_ATOMIC);
  if (pDpcTaskQueue) {
    memset(pDpcTaskQueue, 0, sizeof(*pDpcTaskQueue));
    pDpcTaskQueue->pfnDpc       = pfnDpc;
    pDpcTaskQueue->pDpcContext  = pDpcContext;
    pDpcTaskQueue->pContext     = pContext;
    pDpcTaskQueue->pArg1        = pArg1;
    pDpcTaskQueue->pArg2        = pArg2;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    pDpcTaskQueue->m_tq_struct.routine  = osalinux_call_dpc;
    pDpcTaskQueue->m_tq_struct.data     = pDpcTaskQueue;
    schedule_task(&pDpcTaskQueue->m_tq_struct);
#else
    INIT_WORK(&pDpcTaskQueue->m_tq_struct, osalinux_call_dpc, pDpcTaskQueue);
    schedule_work(&pDpcTaskQueue->m_tq_struct);
#endif
//
//    queue_task(&pDpcTaskQueue->m_tq_struct, &tq_immediate);
//    mark_bh(IMMEDIATE_BH);
//
  } else {
    OSA_EPRINTF(("osalinux_request_dpc:: out of memory!\n"));
  }
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
void osalinux_save_flags_cli(unsigned long *flags) {
 spin_lock_irqsave(&cli_lock, *flags);
}

void osalinux_restore_flags(unsigned long flags) {
 spin_unlock_irqrestore(&cli_lock, flags);
}
#endif

/******************************************************************************
******************************************************************************/

